__all__ = [
    "RetryPolicy",
    "ExceptionRule",
    "ResponseRule",
    "RULE_EXCEPTION_RETRY_TRANSPORT_ERRORS",
    "RULE_RESPONSE_RETRY_SERVER_ERRORS",
    "RULE_RESPONSE_RETRY_UNABLE_TO_LOCK_ROW",
    "RULE_RESPONSE_RETRY_TOO_MANY_REQUESTS",
    "RULE_RESPONSE_RETRY_REQUEST_LIMIT_EXCEEDED",
    "POLICY_DEFAULT",
]

from httpx import TimeoutException, TransportError

from .policy import RetryPolicy
from .rules import ExceptionRule, ResponseRule

RULE_EXCEPTION_RETRY_TRANSPORT_ERRORS = ExceptionRule(
    TransportError,
    lambda exc: not isinstance(exc, TimeoutException),
    max_retries=3,
)
RULE_RESPONSE_RETRY_SERVER_ERRORS = ResponseRule(
    lambda response: response.status_code >= 500,
    max_retries=3,
)
RULE_RESPONSE_RETRY_UNABLE_TO_LOCK_ROW = ResponseRule(
    lambda response: "UNABLE_TO_LOCK_ROW" in response.text,
    max_retries=3,
)
RULE_RESPONSE_RETRY_TOO_MANY_REQUESTS = ResponseRule(
    lambda response: response.status_code == 429,
    max_retries=3,
)
RULE_RESPONSE_RETRY_REQUEST_LIMIT_EXCEEDED = ResponseRule(
    lambda response: "REQUEST_LIMIT_EXCEEDED" in response.text,
    max_retries=3,
)

POLICY_DEFAULT = RetryPolicy(
    response_rules=[
        RULE_RESPONSE_RETRY_SERVER_ERRORS,
        RULE_RESPONSE_RETRY_UNABLE_TO_LOCK_ROW,
        RULE_RESPONSE_RETRY_TOO_MANY_REQUESTS,
        RULE_RESPONSE_RETRY_REQUEST_LIMIT_EXCEEDED,
    ],
    exception_rules=[RULE_EXCEPTION_RETRY_TRANSPORT_ERRORS],
)



import asyncio
import logging

from abc import ABC, abstractmethod
from typing import final

from httpx import AsyncClient

from aiosalesforce.events import EventBus
from aiosalesforce.retries import RetryPolicy

logger = logging.getLogger(__name__)


class Auth(ABC):
    """
    Base class for Salesforce authentication.

    """

    def __init__(self) -> None:
        self.__access_token: str | None = None
        self.__lock = asyncio.Lock()

    @final
    async def get_access_token(
        self,
        client: AsyncClient,
        base_url: str,
        version: str,
        event_bus: EventBus,
        retry_policy: RetryPolicy,
        semaphore: asyncio.Semaphore,
    ) -> str:
        """
        Get access token.

        If this is the first time this method is called, it will acquire a new
        access token from Salesforce.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP client.
        base_url : str
            Salesforce base URL.
            E.g., "https://mydomain.my.salesforce.com".
        version : str
            REST API version.
            E.g., "57.0".
        event_bus : aiosalesforce.events.EventBus
            Event bus.
        retry_policy : aiosalesforce.retries.RetryPolicy
            Retry policy.
        semaphore : asyncio.Semaphore
            Semaphore to limit the number of simultaneous requests.

        Returns
        -------
        str
            Access token

        """
        async with self.__lock:
            if self.__access_token is None:
                logger.debug(
                    "Acquiring new access token using %s for %s",
                    self.__class__.__name__,
                    base_url,
                )
                async with semaphore:
                    self.__access_token = await self._acquire_new_access_token(
                        client=client,
                        base_url=base_url,
                        version=version,
                        event_bus=event_bus,
                        retry_policy=retry_policy,
                    )
            elif self.expired:
                logger.debug(
                    "Token expired, refreshing access token using %s for %s",
                    self.__class__.__name__,
                    base_url,
                )
                async with semaphore:
                    self.__access_token = await self._refresh_access_token(
                        client=client,
                        base_url=base_url,
                        version=version,
                        event_bus=event_bus,
                        retry_policy=retry_policy,
                    )
            return self.__access_token

    @final
    async def refresh_access_token(
        self,
        client: AsyncClient,
        base_url: str,
        version: str,
        event_bus: EventBus,
        retry_policy: RetryPolicy,
        semaphore: asyncio.Semaphore,
    ) -> str:
        """
        Refresh the access token.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP client.
        base_url : str
            Salesforce base URL.
            E.g., "https://mydomain.my.salesforce.com".
        version : str
            REST API version.
            E.g., "57.0".
        event_bus : aiosalesforce.events.EventBus
            Event bus.
        retry_policy : aiosalesforce.retries.RetryPolicy
            Retry policy.
        semaphore : asyncio.Semaphore
            Semaphore to limit the number of simultaneous requests.

        Returns
        -------
        str
            Access token

        """
        if self.__access_token is None:
            raise RuntimeError("No access token to refresh")
        token_before_refresh = self.__access_token
        async with self.__lock:
            if self.__access_token == token_before_refresh:
                logger.debug(
                    "Refreshing access token using %s for %s",
                    self.__class__.__name__,
                    base_url,
                )
                async with semaphore:
                    self.__access_token = await self._refresh_access_token(
                        client=client,
                        base_url=base_url,
                        version=version,
                        event_bus=event_bus,
                        retry_policy=retry_policy,
                    )
            return self.__access_token

    @abstractmethod
    async def _acquire_new_access_token(
        self,
        client: AsyncClient,
        base_url: str,
        version: str,
        event_bus: EventBus,
        retry_policy: RetryPolicy,
    ) -> str:
        """
        Acquire a new access token from Salesforce.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP client.
        base_url : str
            Salesforce base URL.
            E.g., "https://mydomain.my.salesforce.com".
        version : str
            REST API version.
            E.g., "57.0".
        event_bus : aiosalesforce.events.EventBus
            Event bus.
        retry_policy : aiosalesforce.retries.RetryPolicy
            Retry policy.

        Returns
        -------
        str
            Access token

        """

    async def _refresh_access_token(
        self,
        client: AsyncClient,
        base_url: str,
        version: str,
        event_bus: EventBus,
        retry_policy: RetryPolicy,
    ) -> str:
        """
        Refresh the access token.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP client.
        base_url : str
            Salesforce base URL.
            E.g., "https://mydomain.my.salesforce.com".
        version : str
            REST API version.
            E.g., "57.0".
        event_bus : aiosalesforce.events.EventBus
            Event bus.
        retry_policy : aiosalesforce.retries.RetryPolicy
            Retry policy.

        Returns
        -------
        str
            Access token

        """
        return await self._acquire_new_access_token(
            client=client,
            base_url=base_url,
            version=version,
            event_bus=event_bus,
            retry_policy=retry_policy,
        )

    @property
    def expired(self) -> bool:
        """True if the access token is expired."""
        if self.__access_token is None:  # pragma: no cover
            raise RuntimeError("Cannot check expiration of a non-existent access token")
        # By default, assumes the access token never expires
        # Salesforce client automatically refreshes the token after 401 response
        return False



import csv
import datetime

from typing import Any, Iterable


class CsvBuffer:
    """Utility class for buffering CSV data."""

    __content: list[bytes]
    __size: int

    def __init__(self) -> None:
        self.__content = []
        self.__size = 0

    def write(self, row: str) -> None:
        """Write a row to the buffer."""
        encoded_row = row.encode("utf-8")
        self.__content.append(encoded_row)
        self.__size += len(encoded_row)

    def pop(self) -> bytes:
        """Remove and return the last row from the buffer."""
        row = self.__content.pop()
        self.__size -= len(row)
        return row

    def flush(self) -> None:
        """Clear the buffer."""
        self.__content = []
        self.__size = 0

    @property
    def n_rows(self) -> int:
        """Number of rows in the buffer (including the header)."""
        return len(self.__content)

    @property
    def size(self) -> int:
        """Size of the buffer in bytes."""
        return self.__size

    @property
    def content(self) -> bytes:
        """Contents of the buffer as a single byte string."""
        return b"".join(self.__content)


def _serialize_value(value: Any) -> str:
    """Serialize a CSV scalar value."""
    match value:
        case True:
            return "true"
        case False:
            return "false"
        case None:
            return ""
        case datetime.datetime():
            # yyyy-MM-ddTHH:mm:ss.SSS+/-HH:mm
            return value.isoformat(timespec="milliseconds")
        case datetime.date():
            return value.strftime(r"%Y-%m-%d")
        case str() | int() | float():
            return str(value)
        case _:
            raise TypeError(f"Invalid value type '{type(value).__name__}'")


def _serialize_dict(data: dict[str, Any]) -> dict[str, str]:
    """Serialize a CSV row dictionary."""
    new_dict = {}
    for key, value in data.items():
        new_key = key
        if isinstance(value, dict):
            if len(value) != 1:
                raise ValueError(
                    f"Dict for '{key}' must have exactly one value, got {len(value)}"
                )
            new_key = f"{key}.{next(iter(value.keys()))}"
            try:
                new_value = _serialize_value(next(iter(value.values())))
            except TypeError as exc:
                raise TypeError(f"Invalid dict value for '{key}'") from exc
        else:
            new_value = _serialize_value(value)
        new_dict[new_key] = new_value
    return new_dict


def serialize_ingest_data(
    data: Iterable[dict[str, Any]],
    fieldnames: list[str] | None = None,
    max_size_bytes: int = 100_000_000,
    max_records: int = 150_000_000,
) -> Iterable[bytes]:
    """
    Serialize data into CSV files for ingestion by Salesforce Bulk API 2.0.

    None or missing values are ignored by Salesforce.
    To set a field in Salesforce to NULL, use the string "#N/A".
    Relationships are represented as nested dictionaries,
    with exactly one key-value pair. E.g. {"Account": {"Name": "Acme"}}
    or {"Custom_Field__r": {"External_Id__c": "123"}.

    Parameters
    ----------
    data : Iterable[dict[str, Any]]
        Sequence of dictionaries, each representing a record.
    fieldnames : list[str], optional
        List of field names, determines order of fields in the CSV file.
        By default field names are inferred from the records. This is slow, so
        if you know the field names in advance, it is recommended to provide them.
        If a record is missing a field, it will be written as an empty string.
        If a record has a field not in `fieldnames`, an error will be raised.
    max_size_bytes : int, optional
        Maximum size of each CSV file in bytes.
        The default of 100MB is recommended by Salesforce recommends.
        This accounts for base64 encoding increases in size by up to 50%.
    max_records : int, optional
        Maximum number of records in each CSV file. By default 150,000,000.
        This corresponds to the maximum number of records in a 24-hour period.

    Yields
    ------
    bytes
        CSV file as a byte string.

    """
    if fieldnames is None:
        _fields: set[str] = set()
        for record in (_serialize_dict(record) for record in data):
            _fields.update(record.keys())
        fieldnames = list(_fields)

    buffer = CsvBuffer()
    writer = csv.DictWriter(
        buffer,
        fieldnames=fieldnames,
        lineterminator="\n",
    )

    carry_over: bytes | None = None
    for row in (_serialize_dict(record) for record in data):
        if buffer.size == 0:
            writer.writeheader()
            if carry_over is not None:
                buffer.write(carry_over.decode("utf-8"))
                carry_over = None
        writer.writerow(row)
        if buffer.size >= max_size_bytes or buffer.n_rows >= max_records:
            carry_over = buffer.pop()
            yield buffer.content
            buffer.flush()

    if buffer.size > 0:
        yield buffer.content


def deserialize_ingest_results(data: bytes) -> list[dict[str, Any]]:
    """
    Deserialize Salesforce Bulk API 2.0 ingest results from CSV.

    Parameters
    ----------
    data : bytes
        CSV file as a byte string.

    Returns
    -------
    list[dict[str, Any]]
        List of records as dictionaries.

    """
    reader = csv.DictReader(data.decode("utf-8").splitlines())
    return list(reader)



import logging

from dataclasses import dataclass
from typing import TYPE_CHECKING

from aiosalesforce.utils import json_dumps, json_loads

if TYPE_CHECKING:
    from .client import Salesforce

logger = logging.getLogger(__name__)


@dataclass
class UpsertResponse:
    id: str
    created: bool


class SobjectClient:
    """
    Salesforce REST API sObject client.

    Parameters
    ----------
    salesforce_client : Salesforce
        Salesforce client.

    """

    salesforce_client: "Salesforce"
    base_url: str
    """Base URL in the format https://[subdomain(s)].my.salesforce.com/services/data/v[version]/sobjects"""

    def __init__(self, salesforce_client: "Salesforce") -> None:
        self.salesforce_client = salesforce_client
        self.base_url = "/".join(
            [
                f"{self.salesforce_client.base_url}",
                "services",
                "data",
                f"v{self.salesforce_client.version}",
                "sobjects",
            ]
        )

    async def create(
        self,
        sobject: str,
        /,
        data: dict | str | bytes | bytearray,
    ) -> str:
        """
        Create a new record.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
            E.g. "Account", "Contact", etc.
        data : dict | str | bytes | bytearray
            Data to create the record with.
            Either a dict or a JSON string/bytes representing a dict.

        Returns
        -------
        str
            ID of the created record.

        """
        response = await self.salesforce_client.request(
            "POST",
            f"{self.base_url}/{sobject}",
            content=json_dumps(data),
            headers={"Content-Type": "application/json"},
        )
        return json_loads(response.content)["id"]

    async def get(
        self,
        sobject: str,
        id_: str,
        /,
        external_id_field: str | None = None,
        fields: list[str] | None = None,
    ) -> dict:
        """
        Get record by ID or external ID.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
            E.g. "Account", "Contact", etc.
        id_ : str
            Salesforce record ID or external ID (if external_id_field is provided).
        external_id_field : str, optional
            External ID field name, by default None.
        fields : list[str], optional
            Fields to get for the record.
            By default returns all fields.

        Returns
        -------
        dict
            _description_
        """
        url = f"{self.base_url}/{sobject}"
        if external_id_field is None:
            url += f"/{id_}"
        else:
            url += f"/{external_id_field}/{id_}"

        params: dict = {}
        if fields is not None:
            params["fields"] = ",".join(fields)

        response = await self.salesforce_client.request("GET", url, params=params)
        return json_loads(response.content)

    async def update(
        self,
        sobject: str,
        id_: str,
        /,
        data: dict | str | bytes | bytearray,
    ) -> None:
        """
        Update record by ID.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
            E.g. "Account", "Contact", etc.
        id_ : str
            Salesforce record ID.
        data : dict | str | bytes | bytearray
            Data to update the record with.
            Either a dict or a JSON string/bytes representing a dict.

        """
        await self.salesforce_client.request(
            "PATCH",
            f"{self.base_url}/{sobject}/{id_}",
            content=json_dumps(data),
            headers={"Content-Type": "application/json"},
        )

    async def delete(
        self,
        sobject: str,
        id_: str,
        /,
        external_id_field: str | None = None,
    ) -> None:
        """
        Delete record by ID.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
            E.g. "Account", "Contact", etc.
        id_ : str
            Salesforce record ID or external ID (if external_id_field is provided).
        external_id_field : str, optional
            External ID field name.
            If not provided, id_ is treated as a record ID.

        """
        url = f"{self.base_url}/{sobject}"
        if external_id_field is None:
            url += f"/{id_}"
        else:
            url += f"/{external_id_field}/{id_}"
        await self.salesforce_client.request("DELETE", url)

    async def upsert(
        self,
        sobject: str,
        id_: str,
        /,
        external_id_field: str,
        data: dict | str | bytes | bytearray,
        validate: bool = True,
    ) -> UpsertResponse:
        """
        Upsert (update if exists, create if not) record by external ID.

        Parameters
        ----------
        sobject : str
            Salesforce object name.
            E.g. "Account", "Contact", etc.
        id_ : str
            Salesforce record external ID.
        external_id_field : str
            External ID field name.
        data : dict | str | bytes | bytearray
            Data to upsert the record with.
            Either a dict or a JSON string/bytes representing a dict.
        validate : bool, default True
            If True, validates the request and removes the external ID field
            from the data if it's present. By default True.
            The reason for this is that Salesforce does not allow
            payload to contain an external ID field when upserting on it.
            Set this to False if you know you data is correct and
            you want to improve performance.

        Returns
        -------
        UpsertResponse
            Dataclass with 'id' and 'created' fields.

        """
        if validate:
            if isinstance(data, (str, bytes, bytearray)):
                data = json_loads(data)
            if not isinstance(data, dict):
                raise TypeError(
                    f"data must be a dict, str, bytes, or bytearray, "
                    f"got {type(data).__name__}"
                )
            try:
                if data[external_id_field] != id_:
                    raise ValueError(
                        f"External ID field '{external_id_field}' in data "
                        f"{data[external_id_field]} does not match "
                        f"the provided external id {id_}"
                    )
                data.pop(external_id_field)
            except KeyError:
                pass

        response = await self.salesforce_client.request(
            "PATCH",
            f"{self.base_url}/{sobject}/{external_id_field}/{id_}",
            content=json_dumps(data),
            headers={"Content-Type": "application/json"},
        )
        response_json = json_loads(response.content)
        return UpsertResponse(
            id=response_json["id"],
            created=response_json["created"],
        )



import asyncio
import inspect

from typing import Awaitable, Callable, Generic, TypeVar

from httpx import Response

from aiosalesforce.exceptions import SalesforceError

E = TypeVar("E", bound=Exception)


class ResponseRule:
    """
    Rule for deciding if a request should be retried based its response.

    Parameters
    ----------
    func : Callable[[Response], Awaitable[bool] | bool]
        Function or coroutine to determine if the request should be retried.
    max_retries : int, optional
        Maximum number of retries. By default 3.

    """

    func: Callable[[Response], Awaitable[bool] | bool]
    max_retries: int

    def __init__(
        self,
        func: Callable[[Response], Awaitable[bool] | bool],
        /,
        max_retries: int = 3,
    ) -> None:
        self.func = func
        self.max_retries = max_retries

    async def should_retry(self, response: Response) -> bool:
        """
        Determine if the request should be retried.

        Parameters
        ----------
        response : Response
            Response from the request.

        Returns
        -------
        bool
            True if the request should be retried, False otherwise.

        """
        if inspect.iscoroutinefunction(self.func):
            return await self.func(response)
        else:
            return await asyncio.to_thread(
                self.func,  # type: ignore
                response,
            )


class ExceptionRule(Generic[E]):
    """
    Rule for deciding if a request should be retried after an exception.

    Parameters
    ----------
    exc_type : type[Exception]
        Type of exception to retry.
    func : Callable[[Exception], Awaitable[bool]  |  bool] | None, optional
        Function or coroutine to determine if the request should be retried.
        By default the provided exception is always retried.
    max_retries : int, optional
        Maximum number of retries. By default 3.

    """

    exception_type: type[E]
    func: Callable[[E], Awaitable[bool] | bool]
    max_retries: int

    def __init__(
        self,
        exc_type: type[E],
        func: Callable[[E], Awaitable[bool] | bool] | None = None,
        /,
        max_retries: int = 3,
    ) -> None:
        if issubclass(exc_type, SalesforceError):
            raise ValueError(
                "aiosalesforce exceptions cannot be retried by aiosalesforce because "
                "they are raised at the end of the request lifecycle - after all "
                "retries have been exhausted. This could lead to an infinite loop. "
                "If you need to retry aiosalesforce exceptions, consider using "
                "ResponseRule instead."
            )
        if exc_type is Exception:
            raise ValueError("Retrying built-in Exception is not allowed.")
        self.exception_type = exc_type
        self.func = func or (lambda _: True)
        self.max_retries = max_retries

    async def should_retry(self, exc: E) -> bool:
        """
        Determine if the request should be retried.

        Parameters
        ----------
        exc : Exception
            Exception from the request.

        Returns
        -------
        bool
            True if the request should be retried, False otherwise.

        """
        if not isinstance(exc, self.exception_type):
            return False
        if inspect.iscoroutinefunction(self.func):
            return await self.func(exc)
        else:
            return await asyncio.to_thread(
                self.func,  # type: ignore
                exc,
            )



from typing import NoReturn

from httpx import Response

from aiosalesforce.utils import json_loads


class SalesforceWarning(Warning):
    """Base class for all Salesforce warnings."""


class SalesforceError(Exception):
    """Base class for all Salesforce errors."""

    response: Response | None
    error_code: str | None
    error_message: str | None

    def __init__(
        self,
        message: str,
        response: Response | None = None,
        error_code: str | None = None,
        error_message: str | None = None,
    ) -> None:
        super().__init__(message)
        self.response = response
        self.error_code = error_code
        self.error_message = error_message


class MoreThanOneRecordError(SalesforceError):
    """Raised when more than one record is found by external ID."""


class AuthenticationError(SalesforceError):
    """Raised when authentication fails."""


class AuthorizationError(SalesforceError):
    """Raised when user has insufficient permissions to perform an action."""


class RequestLimitExceededError(SalesforceError):
    """Raised when org REST API request limit is exceeded."""


class NotFoundError(SalesforceError):
    """Raised when a resource is not found."""


class ServerError(SalesforceError):
    """Base class for 5xx errors."""


def raise_salesforce_error(response: Response) -> NoReturn:
    """
    Given an HTTP response, raise an appropriate SalesforceError.

    https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/errorcodes.htm

    Parameters
    ----------
    response : httpx.Response
        HTTP response.

    Raises
    ------
    SalesforceError

    """
    try:
        response_json = json_loads(response.content)
        error_code = response_json[0]["errorCode"]
        error_message = response_json[0]["message"]
    except Exception:
        error_code = None
        error_message = response.text

    exc_class: type[SalesforceError]
    match (response.status_code, error_code):
        case (300, _):
            exc_class = MoreThanOneRecordError
            try:
                records = [f"  {record}" for record in json_loads(response.content)]
            except Exception as exc:
                records = [f"  Failed to parse response: {exc}"]
            if error_code is None:
                error_message = "\n".join(
                    [
                        "More than one record found for external ID.",
                        f"{response.url}",
                        *records,
                    ]
                )
        case (_, "REQUEST_LIMIT_EXCEEDED"):
            exc_class = RequestLimitExceededError
        case (403, _):
            exc_class = AuthorizationError
        case (404, _):
            exc_class = NotFoundError
        case (status_code, _) if status_code >= 500:
            exc_class = ServerError
        case _:
            exc_class = SalesforceError
    raise exc_class(
        f"[{error_code}] {error_message}" if error_code else error_message,
        response=response,
        error_code=error_code,
        error_message=error_message,
    )



import asyncio
import random
import time

from typing import TypedDict

from httpx import Response

from .rules import ExceptionRule, ResponseRule


class RetryCounts(TypedDict):
    total: int
    exception: dict[ExceptionRule, int]
    response: dict[ResponseRule, int]


class RetryContext:
    """
    Context for retrying a particular request.

    Parameters
    ----------
    response_rules : list[ResponseRule]
        Rules for retrying requests based on their responses.
    exception_rules : list[ExceptionRule]
        Rules for retrying requests after an exception.
    max_retries : int
        Maximum total number of retries.

    """

    __slots__ = (
        "response_rules",
        "exception_rules",
        "max_retries",
        "timeout",
        "start",
        "retry_count",
    )

    response_rules: list[ResponseRule]
    exception_rules: list[ExceptionRule]
    max_retries: int
    timeout: float
    start: float
    retry_count: RetryCounts

    def __init__(
        self,
        response_rules: list[ResponseRule],
        exception_rules: list[ExceptionRule],
        max_retries: int,
        timeout: float,
    ) -> None:
        self.response_rules = response_rules
        self.exception_rules = exception_rules
        self.max_retries = max_retries
        self.timeout = timeout

        self.start = time.time()
        self.retry_count = {
            "total": 0,
            "response": {rule: 0 for rule in self.response_rules},
            "exception": {rule: 0 for rule in self.exception_rules},
        }

    async def should_retry(self, value: Response | Exception) -> bool:
        """
        Determine if the request should be retried.

        Parameters
        ----------
        value : Response | Exception
            Response or Exception from the request.

        Returns
        -------
        bool
            True if the request should be retried, False otherwise.

        """
        if (
            self.retry_count["total"] >= self.max_retries
            or time.time() - self.start > self.timeout
        ):
            return False
        condition: bool = False
        if isinstance(value, Response):
            condition = await self.__evaluate_response_rules(value)
        elif isinstance(value, Exception):
            condition = await self.__evaluate_exception_rules(value)
        else:
            raise TypeError(  # pragma: no cover
                "Value must be a Response or an Exception"
            )
        if condition:
            self.retry_count["total"] += 1
        return condition

    async def __evaluate_response_rules(self, response: Response) -> bool:
        for rule in self.response_rules:
            if self.retry_count["response"][rule] >= rule.max_retries:
                return False
            if await rule.should_retry(response):
                self.retry_count["response"][rule] += 1
                return True
        return False

    async def __evaluate_exception_rules(self, exception: Exception) -> bool:
        for rule in self.exception_rules:
            if self.retry_count["exception"][rule] >= rule.max_retries:
                return False
            if await rule.should_retry(exception):
                self.retry_count["exception"][rule] += 1
                return True
        return False


class RetryPolicy:
    """
    Policy for retrying requests.

    Parameters
    ----------
    response_rules : list[ResponseRule], optional
        Rules for retrying requests based on their responses.
    exception_rules : list[ExceptionRule], optional
        Rules for retrying requests after an exception.
    max_retries : int, optional
        Maximum total number of retries. By default 3.
    timeout : float, optional
        Maximum time to retry. By default 60 seconds.
        Timeout is best effort and may exceed the specified time
        by up to `backoff_max` seconds.
    backoff_base : float, optional
        Base time to sleep between retries. By default 0.5 seconds.
    backoff_factor : float, optional
        Factor to increase sleep time between retries. By default 2.
    backoff_max : float, optional
        Maximum time to sleep between retries. By default 10 seconds.
    backoff_jitter : bool, optional
        If True, adds jitter to sleep time. By default True.

    """

    response_rules: list[ResponseRule]
    exception_rules: list[ExceptionRule]
    max_retries: int
    timeout: float
    backoff_base: float
    backoff_factor: float
    backoff_max: float
    backoff_jitter: bool

    def __init__(
        self,
        response_rules: list[ResponseRule] | None = None,
        exception_rules: list[ExceptionRule] | None = None,
        max_retries: int = 3,
        timeout: float = 60,
        backoff_base: float = 0.5,
        backoff_factor: float = 2,
        backoff_max: float = 10,
        backoff_jitter: bool = True,
    ) -> None:
        self.exception_rules = exception_rules or []
        self.response_rules = response_rules or []
        self.max_retries = max_retries
        self.timeout = timeout
        self.backoff_base = backoff_base
        self.backoff_factor = backoff_factor
        self.backoff_max = backoff_max
        self.backoff_jitter = backoff_jitter

    def create_context(self) -> RetryContext:
        """
        Create a new retry context.

        Retry context is used to track the state of a retry operation.

        """
        return RetryContext(
            response_rules=self.response_rules,
            exception_rules=self.exception_rules,
            max_retries=self.max_retries,
            timeout=self.timeout,
        )

    async def sleep(self, attempt: int) -> None:
        """
        Sleep between retries.

        Calculated for current attempt using the configured backoff policy.

        Parameters
        ----------
        attempt : int
            Current retry attempt. First attempt is 0.

        """
        sleep_time = min(
            self.backoff_base * (self.backoff_factor**attempt),
            self.backoff_max,
        )
        if self.backoff_jitter:
            sleep_time = random.uniform(0, sleep_time)  # noqa: S311
        await asyncio.sleep(sleep_time)



__all__ = [
    "EventBus",
    "BulkApiBatchConsumptionEvent",
    "Event",
    "RequestEvent",
    "ResponseEvent",
    "RestApiCallConsumptionEvent",
    "RetryEvent",
]

from .event_bus import EventBus
from .events import (
    BulkApiBatchConsumptionEvent,
    Event,
    RequestEvent,
    ResponseEvent,
    RestApiCallConsumptionEvent,
    RetryEvent,
)



import asyncio
import itertools

from httpx import AsyncClient

from aiosalesforce.auth.base import Auth
from aiosalesforce.events import (
    EventBus,
    RequestEvent,
    ResponseEvent,
    RestApiCallConsumptionEvent,
    RetryEvent,
)
from aiosalesforce.exceptions import AuthenticationError
from aiosalesforce.retries import RetryPolicy
from aiosalesforce.utils import json_loads


class ClientCredentialsFlow(Auth):
    """
    Authenticate using the OAuth 2.0 Client Credentials Flow.

    https://help.salesforce.com/s/articleView?id=sf.remoteaccess_oauth_client_credentials_flow.htm&type=5

    Parameters
    ----------
    client_id : str
        Client ID.
    client_secret : str
        Client secret.

    """

    def __init__(self, client_id: str, client_secret: str) -> None:
        super().__init__()
        self.client_id = client_id
        self.client_secret = client_secret

    async def _acquire_new_access_token(
        self,
        client: AsyncClient,
        base_url: str,
        version: str,
        event_bus: EventBus,
        retry_policy: RetryPolicy,
    ) -> str:
        del version  # not used in this flow (this line is for linter)
        request = client.build_request(
            "POST",
            f"{base_url}/services/oauth2/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
            },
            data={
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            },
        )
        await event_bus.publish_event(RequestEvent(type="request", request=request))
        retry_context = retry_policy.create_context()
        for attempt in itertools.count():
            try:
                response = await client.send(request)
            except Exception as exc:
                if await retry_context.should_retry(exc):
                    await asyncio.gather(
                        retry_policy.sleep(attempt),
                        event_bus.publish_event(
                            RetryEvent(
                                type="retry",
                                attempt=attempt + 1,
                                request=request,
                                exception=exc,
                            )
                        ),
                    )
                    continue
                raise
            await event_bus.publish_event(
                RestApiCallConsumptionEvent(
                    type="rest_api_call_consumption",
                    response=response,
                    count=1,
                )
            )
            if response.is_success:
                break
            if await retry_context.should_retry(response):
                await asyncio.gather(
                    retry_policy.sleep(attempt),
                    event_bus.publish_event(
                        RetryEvent(
                            type="retry",
                            attempt=attempt + 1,
                            request=request,
                            response=response,
                        )
                    ),
                )
                continue
            break
        if not response.is_success:
            try:
                response_json = json_loads(response.content)
                error_code = response_json["error"]
                error_message = response_json["error_description"]
            except Exception:  # pragma: no cover
                error_code = None
                error_message = response.text
            raise AuthenticationError(
                f"[{error_code}] {error_message}" if error_code else error_message,
                response=response,
                error_code=error_code,
                error_message=error_message,
            )
        await event_bus.publish_event(ResponseEvent(type="response", response=response))
        return json_loads(response.content)["access_token"]



import asyncio
import inspect

from typing import Awaitable, Callable, TypeAlias

from .events import Event

CallbackType: TypeAlias = Callable[[Event], Awaitable[None] | None]


class EventBus:
    """
    Event bus used to dispatch events to subscribed callbacks.

    Parameters
    ----------
    callbacks : list[Callable[[Event], Awaitable[None] | None]], optional
        List of callbacks to subscribe to the event bus.

    """

    _callbacks: set[CallbackType]

    def __init__(self, callbacks: list[CallbackType] | None = None) -> None:
        self._callbacks = set()
        for callback in callbacks or []:
            self.subscribe_callback(callback)

    def subscribe_callback(self, callback: CallbackType) -> None:
        """
        Subscribe a callback to the event bus.

        Parameters
        ----------
        callback : CallbackType
            Function to be called when an event is published.

        """
        self._callbacks.add(callback)

    def unsubscribe_callback(self, callback: CallbackType) -> None:
        """
        Unsubscribe a callback from the event bus.

        Parameters
        ----------
        callback : CallbackType
            Function to be unsubscribed.

        """
        self._callbacks.discard(callback)

    @staticmethod
    async def __dispatch_event_to_callback(
        callback: CallbackType, event: Event
    ) -> None:
        if inspect.iscoroutinefunction(callback):
            await callback(event)
        else:
            await asyncio.to_thread(callback, event)

    async def publish_event(self, event: Event) -> None:
        """Publish an event and dispatch it to all subscribed callbacks."""
        async with asyncio.TaskGroup() as tg:
            for callback in self._callbacks:
                tg.create_task(self.__dispatch_event_to_callback(callback, event))



__version__ = "0.4.0"

__all__ = [
    "Salesforce",
    "BulkApiBatchConsumptionEvent",
    "RequestEvent",
    "ResponseEvent",
    "RestApiCallConsumptionEvent",
    "RetryEvent",
    "ExceptionRule",
    "ResponseRule",
    "RetryPolicy",
    "format_soql",
]

from .client import Salesforce
from .events import (
    BulkApiBatchConsumptionEvent,
    RequestEvent,
    ResponseEvent,
    RestApiCallConsumptionEvent,
    RetryEvent,
)
from .retries import (
    ExceptionRule,
    ResponseRule,
    RetryPolicy,
)
from .utils import format_soql
